@@ -141,6 +141,10 @@ pub struct Component {
     /// stored in two different locations).
     pub num_runtime_memories: u32,
 
+    /// The number of runtime tables (maximum `RuntimeTableIndex`) needed to
+    /// instantiate this component. See notes on `num_runtime_memories`.
+    pub num_runtime_tables: u32,
+
     /// The number of runtime reallocs (maximum `RuntimeReallocIndex`) needed to
     /// instantiate this component.
     ///
@@ -215,10 +219,11 @@ impl Component {
     }
 }
 
-/// GlobalInitializer instructions to get processed when instantiating a component
+/// GlobalInitializer instructions to get processed when instantiating a
+/// component.
 ///
-/// The variants of this enum are processed during the instantiation phase of
-/// a component in-order from front-to-back. These are otherwise emitted as a
+/// The variants of this enum are processed during the instantiation phase of a
+/// component in-order from front-to-back. These are otherwise emitted as a
 /// component is parsed and read and translated.
 //
 // FIXME(#2639) if processing this list is ever a bottleneck we could
@@ -257,11 +262,10 @@ pub enum GlobalInitializer {
     /// A core wasm linear memory is going to be saved into the
     /// `VMComponentContext`.
     ///
-    /// This instruction indicates that the `index`th core wasm linear memory
-    /// needs to be extracted from the `export` specified, a pointer to a
-    /// previously created module instance, and stored into the
-    /// `VMComponentContext` at the `index` specified. This lowering is then
-    /// used in the future by pointers from `CanonicalOptions`.
+    /// This instruction indicates that a core wasm linear memory needs to be
+    /// extracted from the `export` and stored into the `VMComponentContext` at
+    /// the `index` specified. This lowering is then used in the future by
+    /// pointers from `CanonicalOptions`.
     ExtractMemory(ExtractMemory),
 
     /// Same as `ExtractMemory`, except it's extracting a function pointer to be
@@ -276,14 +280,24 @@ pub enum GlobalInitializer {
     /// used as a `post-return` function.
     ExtractPostReturn(ExtractPostReturn),
 
+    /// A core wasm table is going to be saved into the `VMComponentContext`.
+    ///
+    /// This instruction indicates that s core wasm table needs to be extracted
+    /// from its `export` and stored into the `VMComponentContext` at the
+    /// `index` specified. During this extraction, we will also capture the
+    /// table's containing instance pointer to access the table at runtime. This
+    /// extraction is useful for `thread.spawn_indirect`.
+    ExtractTable(ExtractTable),
+
     /// Declares a new defined resource within this component.
     ///
     /// Contains information about the destructor, for example.
     Resource(Resource),
 }
 
-/// Metadata for extraction of a memory of what's being extracted and where it's
-/// going.
+/// Metadata for extraction of a memory; contains what's being extracted (the
+/// memory at `export`) and where it's going (the `index` within a
+/// `VMComponentContext`).
 #[derive(Debug, Serialize, Deserialize)]
 pub struct ExtractMemory {
     /// The index of the memory being defined.
@@ -319,6 +333,15 @@ pub struct ExtractPostReturn {
     pub def: CoreDef,
 }
 
+/// Metadata for extraction of a table.
+#[derive(Debug, Serialize, Deserialize)]
+pub struct ExtractTable {
+    /// The index of the table being defined in a `VMComponentContext`.
+    pub index: RuntimeTableIndex,
+    /// Where this table is being extracted from.
+    pub export: CoreExport<TableIndex>,
+}
+
 /// Different methods of instantiating a core wasm module.
 #[derive(Debug, Serialize, Deserialize)]
 pub enum InstantiateModule {
@@ -488,7 +511,7 @@ pub struct CanonicalOptions {
 
 /// Possible encodings of strings within the component model.
 #[derive(Debug, Clone, Copy, Serialize, Deserialize, PartialEq, Eq, Hash)]
-#[allow(missing_docs, reason = "self-describing variants")]
+#[expect(missing_docs, reason = "self-describing variants")]
 pub enum StringEncoding {
     Utf8,
     Utf16,
@@ -515,7 +538,7 @@ impl StringEncoding {
 ///
 /// Note that each transcoding operation may have a unique signature depending
 /// on the precise operation.
-#[allow(missing_docs, reason = "self-describing variants")]
+#[expect(missing_docs, reason = "self-describing variants")]
 #[derive(Debug, Copy, Clone, Hash, Eq, PartialEq)]
 pub enum Transcode {
     Copy(FixedEncoding),
@@ -571,7 +594,7 @@ impl Transcode {
 }
 
 #[derive(Debug, Copy, Clone, Hash, Eq, PartialEq, Serialize, Deserialize)]
-#[allow(missing_docs, reason = "self-describing variants")]
+#[expect(missing_docs, reason = "self-describing variants")]
 pub enum FixedEncoding {
     Utf8,
     Utf16,
@@ -679,9 +702,9 @@ pub enum Trampoline {
     /// Same as `ResourceNew`, but for the `resource.drop` intrinsic.
     ResourceDrop(TypeResourceTableIndex),
 
-    /// A `task.backpressure` intrinsic, which tells the host to enable or
+    /// A `backpressure.set` intrinsic, which tells the host to enable or
     /// disable backpressure for the caller's instance.
-    TaskBackpressure {
+    BackpressureSet {
         /// The specific component instance which is calling the intrinsic.
         instance: RuntimeComponentInstanceIndex,
     },
@@ -689,11 +712,24 @@ pub enum Trampoline {
     /// A `task.return` intrinsic, which returns a result to the caller of a
     /// lifted export function.  This allows the callee to continue executing
     /// after returning a result.
-    TaskReturn,
+    TaskReturn {
+        /// Tuple representing the result types this intrinsic accepts.
+        results: TypeTupleIndex,
 
-    /// A `task.wait` intrinsic, which waits for at least one outstanding async
-    /// task/stream/future to make progress, returning the first such event.
-    TaskWait {
+        /// The canonical ABI options specified for this intrinsic.
+        options: CanonicalOptions,
+    },
+
+    /// A `waitable-set.new` intrinsic.
+    WaitableSetNew {
+        /// The specific component instance which is calling the intrinsic.
+        instance: RuntimeComponentInstanceIndex,
+    },
+
+    /// A `waitable-set.wait` intrinsic, which waits for at least one
+    /// outstanding async task/stream/future to make progress, returning the
+    /// first such event.
+    WaitableSetWait {
         /// The specific component instance which is calling the intrinsic.
         instance: RuntimeComponentInstanceIndex,
         /// If `true`, indicates the caller instance maybe reentered.
@@ -702,10 +738,10 @@ pub enum Trampoline {
         memory: RuntimeMemoryIndex,
     },
 
-    /// A `task.poll` intrinsic, which checks whether any outstanding async
-    /// task/stream/future has made progress.  Unlike `task.wait`, this does not
-    /// block and may return nothing if no such event has occurred.
-    TaskPoll {
+    /// A `waitable-set.poll` intrinsic, which checks whether any outstanding
+    /// async task/stream/future has made progress.  Unlike `task.wait`, this
+    /// does not block and may return nothing if no such event has occurred.
+    WaitableSetPoll {
         /// The specific component instance which is calling the intrinsic.
         instance: RuntimeComponentInstanceIndex,
         /// If `true`, indicates the caller instance maybe reentered.
@@ -714,9 +750,21 @@ pub enum Trampoline {
         memory: RuntimeMemoryIndex,
     },
 
-    /// A `task.yield` intrinsic, which yields control to the host so that other
+    /// A `waitable-set.drop` intrinsic.
+    WaitableSetDrop {
+        /// The specific component instance which is calling the intrinsic.
+        instance: RuntimeComponentInstanceIndex,
+    },
+
+    /// A `waitable.join` intrinsic.
+    WaitableJoin {
+        /// The specific component instance which is calling the intrinsic.
+        instance: RuntimeComponentInstanceIndex,
+    },
+
+    /// A `yield` intrinsic, which yields control to the host so that other
     /// tasks are able to make progress, if any.
-    TaskYield {
+    Yield {
         /// If `true`, indicates the caller instance maybe reentered.
         async_: bool,
     },
@@ -738,6 +786,7 @@ pub enum Trampoline {
     StreamRead {
         /// The table index for the specific `stream` type and caller instance.
         ty: TypeStreamTableIndex,
+
         /// Any options (e.g. string encoding) to use when storing values to
         /// memory.
         options: CanonicalOptions,
@@ -797,6 +846,7 @@ pub enum Trampoline {
     FutureRead {
         /// The table index for the specific `future` type and caller instance.
         ty: TypeFutureTableIndex,
+
         /// Any options (e.g. string encoding) to use when storing values to
         /// memory.
         options: CanonicalOptions,
@@ -890,12 +940,23 @@ pub enum Trampoline {
     /// Same as `ResourceEnterCall` except for when exiting a call.
     ResourceExitCall,
 
-    /// An intrinsic used by FACT-generated modules to begin a call to an
+    /// An intrinsic used by FACT-generated modules to begin a call involving a
+    /// sync-lowered import and async-lifted export.
+    SyncEnterCall,
+
+    /// An intrinsic used by FACT-generated modules to complete a call involving
+    /// a sync-lowered import and async-lifted export.
+    SyncExitCall {
+        /// The callee's callback function, if any.
+        callback: Option<RuntimeCallbackIndex>,
+    },
+
+    /// An intrinsic used by FACT-generated modules to begin a call involving an
     /// async-lowered import function.
     AsyncEnterCall,
 
-    /// An intrinsic used by FACT-generated modules to complete a call to an
-    /// async-lowered import function.
+    /// An intrinsic used by FACT-generated modules to complete a call involving
+    /// an async-lowered import function.
     ///
     /// Note that `AsyncEnterCall` and `AsyncExitCall` could theoretically be
     /// combined into a single `AsyncCall` intrinsic, but we separate them to
@@ -955,11 +1016,14 @@ impl Trampoline {
             ResourceNew(i) => format!("component-resource-new[{}]", i.as_u32()),
             ResourceRep(i) => format!("component-resource-rep[{}]", i.as_u32()),
             ResourceDrop(i) => format!("component-resource-drop[{}]", i.as_u32()),
-            TaskBackpressure { .. } => format!("task-backpressure"),
-            TaskReturn => format!("task-return"),
-            TaskWait { .. } => format!("task-wait"),
-            TaskPoll { .. } => format!("task-poll"),
-            TaskYield { .. } => format!("task-yield"),
+            BackpressureSet { .. } => format!("backpressure-set"),
+            TaskReturn { .. } => format!("task-return"),
+            WaitableSetNew { .. } => format!("waitable-set-new"),
+            WaitableSetWait { .. } => format!("waitable-set-wait"),
+            WaitableSetPoll { .. } => format!("waitable-set-poll"),
+            WaitableSetDrop { .. } => format!("waitable-set-drop"),
+            WaitableJoin { .. } => format!("waitable-join"),
+            Yield { .. } => format!("yield"),
             SubtaskDrop { .. } => format!("subtask-drop"),
             StreamNew { .. } => format!("stream-new"),
             StreamRead { .. } => format!("stream-read"),
@@ -982,6 +1046,8 @@ impl Trampoline {
             ResourceTransferBorrow => format!("component-resource-transfer-borrow"),
             ResourceEnterCall => format!("component-resource-enter-call"),
             ResourceExitCall => format!("component-resource-exit-call"),
+            SyncEnterCall => format!("component-sync-enter-call"),
+            SyncExitCall { .. } => format!("component-sync-exit-call"),
             AsyncEnterCall => format!("component-async-enter-call"),
             AsyncExitCall { .. } => format!("component-async-exit-call"),
             FutureTransfer => format!("future-transfer"),
