@@ -3,8 +3,8 @@ use crate::prelude::*;
 use crate::ScopeVec;
 use crate::{
     EngineOrModuleTypeIndex, EntityIndex, ModuleEnvironment, ModuleInternedTypeIndex,
-    ModuleTranslation, ModuleTypesBuilder, PrimaryMap, Tunables, TypeConvert, WasmHeapType,
-    WasmResult, WasmValType,
+    ModuleTranslation, ModuleTypesBuilder, PrimaryMap, TagIndex, Tunables, TypeConvert,
+    WasmHeapType, WasmResult, WasmValType,
 };
 use anyhow::anyhow;
 use anyhow::{bail, Result};
@@ -13,7 +13,7 @@ use std::collections::HashMap;
 use std::mem;
 use wasmparser::component_types::{
     AliasableResourceId, ComponentCoreModuleTypeId, ComponentDefinedTypeId, ComponentEntityType,
-    ComponentFuncTypeId, ComponentInstanceTypeId,
+    ComponentFuncTypeId, ComponentInstanceTypeId, ComponentValType,
 };
 use wasmparser::types::Types;
 use wasmparser::{Chunk, ComponentImportName, Encoding, Parser, Payload, Validator};
@@ -188,23 +188,34 @@ enum LocalInitializer<'data> {
     ResourceRep(AliasableResourceId, ModuleInternedTypeIndex),
     ResourceDrop(AliasableResourceId, ModuleInternedTypeIndex),
 
-    TaskBackpressure {
+    BackpressureSet {
         func: ModuleInternedTypeIndex,
     },
     TaskReturn {
         func: ModuleInternedTypeIndex,
+        result: Option<ComponentValType>,
+        options: LocalCanonicalOptions,
     },
-    TaskWait {
+    WaitableSetNew {
+        func: ModuleInternedTypeIndex,
+    },
+    WaitableSetWait {
         func: ModuleInternedTypeIndex,
         async_: bool,
         memory: MemoryIndex,
     },
-    TaskPoll {
+    WaitableSetPoll {
         func: ModuleInternedTypeIndex,
         async_: bool,
         memory: MemoryIndex,
     },
-    TaskYield {
+    WaitableSetDrop {
+        func: ModuleInternedTypeIndex,
+    },
+    WaitableJoin {
+        func: ModuleInternedTypeIndex,
+    },
+    Yield {
         func: ModuleInternedTypeIndex,
         async_: bool,
     },
@@ -310,6 +321,7 @@ enum LocalInitializer<'data> {
     AliasExportTable(ModuleInstanceIndex, &'data str),
     AliasExportGlobal(ModuleInstanceIndex, &'data str),
     AliasExportMemory(ModuleInstanceIndex, &'data str),
+    AliasExportTag(ModuleInstanceIndex, &'data str),
     AliasComponentExport(ComponentInstanceIndex, &'data str),
     AliasModule(ClosedOverModule),
     AliasComponent(ClosedOverComponent),
@@ -617,48 +629,76 @@ impl<'a, 'data> Translator<'a, 'data> {
                             core_func_index += 1;
                             LocalInitializer::ResourceDrop(resource, ty)
                         }
+                        wasmparser::CanonicalFunction::ResourceDropAsync { resource } => {
+                            let _ = resource;
+                            bail!("support for `resource.drop async` not implemented yet")
+                        }
                         wasmparser::CanonicalFunction::ResourceRep { resource } => {
                             let resource = types.component_any_type_at(resource).unwrap_resource();
                             let ty = self.core_func_signature(core_func_index)?;
                             core_func_index += 1;
                             LocalInitializer::ResourceRep(resource, ty)
                         }
-                        wasmparser::CanonicalFunction::ThreadSpawn { .. }
-                        | wasmparser::CanonicalFunction::ThreadAvailableParallelism => {
-                            bail!("unsupported intrinsic")
-                        }
-                        wasmparser::CanonicalFunction::TaskBackpressure => {
+                        wasmparser::CanonicalFunction::BackpressureSet => {
                             let core_type = self.core_func_signature(core_func_index)?;
                             core_func_index += 1;
-                            LocalInitializer::TaskBackpressure { func: core_type }
+                            LocalInitializer::BackpressureSet { func: core_type }
                         }
-                        wasmparser::CanonicalFunction::TaskReturn { .. } => {
-                            let core_type = self.core_func_signature(core_func_index)?;
-                            core_func_index += 1;
-                            LocalInitializer::TaskReturn { func: core_type }
-                        }
-                        wasmparser::CanonicalFunction::TaskWait { async_, memory } => {
+                        wasmparser::CanonicalFunction::TaskReturn { result, options } => {
+                            let result = result.map(|ty| match ty {
+                                wasmparser::ComponentValType::Primitive(ty) => {
+                                    ComponentValType::Primitive(ty)
+                                }
+                                wasmparser::ComponentValType::Type(ty) => {
+                                    ComponentValType::Type(types.component_defined_type_at(ty))
+                                }
+                            });
+                            let options = self.canonical_options(&options);
                             let func = self.core_func_signature(core_func_index)?;
                             core_func_index += 1;
-                            LocalInitializer::TaskWait {
+                            LocalInitializer::TaskReturn {
+                                func,
+                                result,
+                                options,
+                            }
+                        }
+                        wasmparser::CanonicalFunction::WaitableSetNew => {
+                            let func = self.core_func_signature(core_func_index)?;
+                            core_func_index += 1;
+                            LocalInitializer::WaitableSetNew { func }
+                        }
+                        wasmparser::CanonicalFunction::WaitableSetWait { async_, memory } => {
+                            let func = self.core_func_signature(core_func_index)?;
+                            core_func_index += 1;
+                            LocalInitializer::WaitableSetWait {
                                 func,
                                 async_,
                                 memory: MemoryIndex::from_u32(memory),
                             }
                         }
-                        wasmparser::CanonicalFunction::TaskPoll { async_, memory } => {
+                        wasmparser::CanonicalFunction::WaitableSetPoll { async_, memory } => {
                             let func = self.core_func_signature(core_func_index)?;
                             core_func_index += 1;
-                            LocalInitializer::TaskPoll {
+                            LocalInitializer::WaitableSetPoll {
                                 func,
                                 async_,
                                 memory: MemoryIndex::from_u32(memory),
                             }
                         }
-                        wasmparser::CanonicalFunction::TaskYield { async_ } => {
+                        wasmparser::CanonicalFunction::WaitableSetDrop => {
                             let func = self.core_func_signature(core_func_index)?;
                             core_func_index += 1;
-                            LocalInitializer::TaskYield { func, async_ }
+                            LocalInitializer::WaitableSetDrop { func }
+                        }
+                        wasmparser::CanonicalFunction::WaitableJoin => {
+                            let func = self.core_func_signature(core_func_index)?;
+                            core_func_index += 1;
+                            LocalInitializer::WaitableJoin { func }
+                        }
+                        wasmparser::CanonicalFunction::Yield { async_ } => {
+                            let func = self.core_func_signature(core_func_index)?;
+                            core_func_index += 1;
+                            LocalInitializer::Yield { func, async_ }
                         }
                         wasmparser::CanonicalFunction::SubtaskDrop => {
                             let func = self.core_func_signature(core_func_index)?;
@@ -770,6 +810,13 @@ impl<'a, 'data> Translator<'a, 'data> {
                             core_func_index += 1;
                             LocalInitializer::ErrorContextDrop { func }
                         }
+                        wasmparser::CanonicalFunction::ContextGet(..)
+                        | wasmparser::CanonicalFunction::ContextSet(..)
+                        | wasmparser::CanonicalFunction::ThreadSpawnRef { .. }
+                        | wasmparser::CanonicalFunction::ThreadSpawnIndirect { .. }
+                        | wasmparser::CanonicalFunction::ThreadAvailableParallelism => {
+                            bail!("unsupported intrinsic")
+                        }
                     };
                     self.result.initializers.push(init);
                 }
@@ -996,9 +1043,10 @@ impl<'a, 'data> Translator<'a, 'data> {
                     let index = GlobalIndex::from_u32(export.index);
                     EntityIndex::Global(index)
                 }
-
-                // doesn't get past validation
-                wasmparser::ExternalKind::Tag => unimplemented!("wasm exceptions"),
+                wasmparser::ExternalKind::Tag => {
+                    let index = TagIndex::from_u32(export.index);
+                    EntityIndex::Tag(index)
+                }
             };
             map.insert(export.name, idx);
         }
@@ -1080,9 +1128,7 @@ impl<'a, 'data> Translator<'a, 'data> {
             wasmparser::ExternalKind::Memory => LocalInitializer::AliasExportMemory(instance, name),
             wasmparser::ExternalKind::Table => LocalInitializer::AliasExportTable(instance, name),
             wasmparser::ExternalKind::Global => LocalInitializer::AliasExportGlobal(instance, name),
-            wasmparser::ExternalKind::Tag => {
-                unimplemented!("wasm exceptions");
-            }
+            wasmparser::ExternalKind::Tag => LocalInitializer::AliasExportTag(instance, name),
         }
     }
 
