@@ -304,6 +304,12 @@ fn fact_import_to_core_def(
         }
         fact::Import::ResourceEnterCall => simple_intrinsic(dfg::Trampoline::ResourceEnterCall),
         fact::Import::ResourceExitCall => simple_intrinsic(dfg::Trampoline::ResourceExitCall),
+        fact::Import::SyncEnterCall => simple_intrinsic(dfg::Trampoline::SyncEnterCall),
+        fact::Import::SyncExitCall { callback } => {
+            simple_intrinsic(dfg::Trampoline::SyncExitCall {
+                callback: callback.clone().map(|v| dfg.callbacks.push(v)),
+            })
+        }
         fact::Import::AsyncEnterCall => simple_intrinsic(dfg::Trampoline::AsyncEnterCall),
         fact::Import::AsyncExitCall {
             callback,
