@@ -204,6 +204,7 @@ struct InlinerFrame<'a> {
     memories: PrimaryMap<MemoryIndex, dfg::CoreExport<EntityIndex>>,
     tables: PrimaryMap<TableIndex, dfg::CoreExport<EntityIndex>>,
     globals: PrimaryMap<GlobalIndex, dfg::CoreExport<EntityIndex>>,
+    tags: PrimaryMap<GlobalIndex, dfg::CoreExport<EntityIndex>>,
     modules: PrimaryMap<ModuleIndex, ModuleDef<'a>>,
 
     // component model index spaces
@@ -670,23 +671,43 @@ impl<'a> Inliner<'a> {
                     .push((*ty, dfg::Trampoline::ResourceDrop(id)));
                 frame.funcs.push(dfg::CoreDef::Trampoline(index));
             }
-            TaskBackpressure { func } => {
+            BackpressureSet { func } => {
                 let index = self.result.trampolines.push((
                     *func,
-                    dfg::Trampoline::TaskBackpressure {
+                    dfg::Trampoline::BackpressureSet {
                         instance: frame.instance,
                     },
                 ));
                 frame.funcs.push(dfg::CoreDef::Trampoline(index));
             }
-            TaskReturn { func } => {
+            TaskReturn {
+                func,
+                result,
+                options,
+            } => {
+                let results = result
+                    .iter()
+                    .map(|ty| types.valtype(frame.translation.types_ref(), ty))
+                    .collect::<Result<_>>()?;
+                let results = types.new_tuple_type(results);
+                let options = self.adapter_options(frame, types, options);
+                let options = self.canonical_options(options);
                 let index = self
                     .result
                     .trampolines
-                    .push((*func, dfg::Trampoline::TaskReturn));
+                    .push((*func, dfg::Trampoline::TaskReturn { results, options }));
                 frame.funcs.push(dfg::CoreDef::Trampoline(index));
             }
-            TaskWait {
+            WaitableSetNew { func } => {
+                let index = self.result.trampolines.push((
+                    *func,
+                    dfg::Trampoline::WaitableSetNew {
+                        instance: frame.instance,
+                    },
+                ));
+                frame.funcs.push(dfg::CoreDef::Trampoline(index));
+            }
+            WaitableSetWait {
                 func,
                 async_,
                 memory,
@@ -695,7 +716,7 @@ impl<'a> Inliner<'a> {
                 let memory = self.result.memories.push(memory);
                 let index = self.result.trampolines.push((
                     *func,
-                    dfg::Trampoline::TaskWait {
+                    dfg::Trampoline::WaitableSetWait {
                         instance: frame.instance,
                         async_: *async_,
                         memory,
@@ -703,7 +724,7 @@ impl<'a> Inliner<'a> {
                 ));
                 frame.funcs.push(dfg::CoreDef::Trampoline(index));
             }
-            TaskPoll {
+            WaitableSetPoll {
                 func,
                 async_,
                 memory,
@@ -712,7 +733,7 @@ impl<'a> Inliner<'a> {
                 let memory = self.result.memories.push(memory);
                 let index = self.result.trampolines.push((
                     *func,
-                    dfg::Trampoline::TaskPoll {
+                    dfg::Trampoline::WaitableSetPoll {
                         instance: frame.instance,
                         async_: *async_,
                         memory,
@@ -720,11 +741,29 @@ impl<'a> Inliner<'a> {
                 ));
                 frame.funcs.push(dfg::CoreDef::Trampoline(index));
             }
-            TaskYield { func, async_ } => {
+            WaitableSetDrop { func } => {
+                let index = self.result.trampolines.push((
+                    *func,
+                    dfg::Trampoline::WaitableSetDrop {
+                        instance: frame.instance,
+                    },
+                ));
+                frame.funcs.push(dfg::CoreDef::Trampoline(index));
+            }
+            WaitableJoin { func } => {
+                let index = self.result.trampolines.push((
+                    *func,
+                    dfg::Trampoline::WaitableJoin {
+                        instance: frame.instance,
+                    },
+                ));
+                frame.funcs.push(dfg::CoreDef::Trampoline(index));
+            }
+            Yield { func, async_ } => {
                 let index = self
                     .result
                     .trampolines
-                    .push((*func, dfg::Trampoline::TaskYield { async_: *async_ }));
+                    .push((*func, dfg::Trampoline::Yield { async_: *async_ }));
                 frame.funcs.push(dfg::CoreDef::Trampoline(index));
             }
             SubtaskDrop { func } => {
@@ -1114,6 +1153,15 @@ impl<'a> Inliner<'a> {
                 );
             }
 
+            AliasExportTag(instance, name) => {
+                frame.tags.push(
+                    match self.core_def_of_module_instance_export(frame, *instance, *name) {
+                        dfg::CoreDef::Export(e) => e,
+                        _ => unreachable!(),
+                    },
+                );
+            }
+
             AliasComponentExport(instance, name) => {
                 match &frame.component_instances[*instance] {
                     // Aliasing an export from an imported instance means that
@@ -1430,6 +1478,7 @@ impl<'a> InlinerFrame<'a> {
             memories: Default::default(),
             tables: Default::default(),
             globals: Default::default(),
+            tags: Default::default(),
 
             component_instances: Default::default(),
             component_funcs: Default::default(),
