@@ -16,6 +16,8 @@ pub struct ModuleConfig {
     // config-to-`wasmtime::Config` translation.
     pub function_references_enabled: bool,
     pub component_model_async: bool,
+    pub component_model_async_builtins: bool,
+    pub component_model_async_stackful: bool,
 }
 
 impl<'a> Arbitrary<'a> for ModuleConfig {
@@ -41,8 +43,8 @@ impl<'a> Arbitrary<'a> for ModuleConfig {
         let _ = config.relaxed_simd_enabled;
         let _ = config.tail_call_enabled;
         let _ = config.extended_const_enabled;
+        let _ = config.gc_enabled;
         config.exceptions_enabled = false;
-        config.gc_enabled = false;
         config.custom_page_sizes_enabled = u.arbitrary()?;
         config.wide_arithmetic_enabled = u.arbitrary()?;
         config.memory64_enabled = u.ratio(1, 20)?;
@@ -62,6 +64,8 @@ impl<'a> Arbitrary<'a> for ModuleConfig {
 
         Ok(ModuleConfig {
             component_model_async: false,
+            component_model_async_builtins: false,
+            component_model_async_stackful: false,
             function_references_enabled: config.gc_enabled,
             config,
         })
