@@ -246,13 +246,27 @@ pub fn instantiate_many(
     config: &generators::Config,
     commands: &[Command],
 ) {
+    log::debug!("instantiate_many: {commands:#?}");
+
     assert!(!config.module_config.config.allow_start_export);
 
     let engine = Engine::new(&config.to_wasmtime()).unwrap();
 
     let modules = modules
         .iter()
-        .filter_map(|bytes| compile_module(&engine, bytes, known_valid, config))
+        .enumerate()
+        .filter_map(
+            |(i, bytes)| match compile_module(&engine, bytes, known_valid, config) {
+                Some(m) => {
+                    log::debug!("successfully compiled module {i}");
+                    Some(m)
+                }
+                None => {
+                    log::debug!("failed to compile module {i}");
+                    None
+                }
+            },
+        )
         .collect::<Vec<_>>();
 
     // If no modules were valid, we're done
@@ -318,7 +332,7 @@ fn compile_module(
                 // when arbitrary table element limits have been exceeded as
                 // there is currently no way to constrain the generated module
                 // table types.
-                let string = e.to_string();
+                let string = format!("{e:?}");
                 if string.contains("minimum element size") {
                     return None;
                 }
@@ -369,8 +383,11 @@ pub fn instantiate_with_dummy(store: &mut Store<StoreLimits>, module: &Module) -
     // Creation of imports can fail due to resource limit constraints, and then
     // instantiation can naturally fail for a number of reasons as well. Bundle
     // the two steps together to match on the error below.
-    let instance =
-        dummy::dummy_linker(store, module).and_then(|l| l.instantiate(&mut *store, module));
+    let linker = dummy::dummy_linker(store, module);
+    if let Err(e) = &linker {
+        log::warn!("failed to create dummy linker: {e:?}");
+    }
+    let instance = linker.and_then(|l| l.instantiate(&mut *store, module));
     unwrap_instance(store, instance)
 }
 
@@ -383,33 +400,32 @@ fn unwrap_instance(
         Err(e) => e,
     };
 
+    log::debug!("failed to instantiate: {e:?}");
+
     // If the instantiation hit OOM for some reason then that's ok, it's
     // expected that fuzz-generated programs try to allocate lots of
     // stuff.
     if store.data().is_oom() {
-        log::debug!("failed to instantiate: OOM");
         return None;
     }
 
     // Allow traps which can happen normally with `unreachable` or a
     // timeout or such
-    if let Some(trap) = e.downcast_ref::<Trap>() {
-        log::debug!("failed to instantiate: {}", trap);
+    if e.is::<Trap>() {
         return None;
     }
 
     let string = e.to_string();
+
     // Currently we instantiate with a `Linker` which can't instantiate
     // every single module under the sun due to using name-based resolution
     // rather than positional-based resolution
     if string.contains("incompatible import type") {
-        log::debug!("failed to instantiate: {}", string);
         return None;
     }
 
     // Also allow failures to instantiate as a result of hitting pooling limits.
     if e.is::<wasmtime::PoolConcurrencyLimitError>() {
-        log::debug!("failed to instantiate: {}", string);
         return None;
     }
 
@@ -507,6 +523,32 @@ pub enum DiffEqResult<T, U> {
     Failed,
 }
 
+fn wasmtime_trap_is_non_deterministic(trap: &Trap) -> bool {
+    match trap {
+        // Allocations being too large for the GC are
+        // implementation-defined.
+        Trap::AllocationTooLarge |
+        // Stack size, and therefore when overflow happens, is
+        // implementation-defined.
+        Trap::StackOverflow => true,
+        _ => false,
+    }
+}
+
+fn wasmtime_error_is_non_deterministic(error: &wasmtime::Error) -> bool {
+    match error.downcast_ref::<Trap>() {
+        Some(trap) => wasmtime_trap_is_non_deterministic(trap),
+
+        // For general, unknown errors, we can't rely on this being
+        // a deterministic Wasm failure that both engines handled
+        // identically, leaving Wasm in identical states. We could
+        // just as easily be hitting engine-specific failures, like
+        // different implementation-defined limits. So simply poison
+        // this execution and move on to the next test.
+        None => true,
+    }
+}
+
 impl<T, U> DiffEqResult<T, U> {
     /// Computes the differential result from executing in two different
     /// engines.
@@ -518,41 +560,37 @@ impl<T, U> DiffEqResult<T, U> {
         match (lhs_result, rhs_result) {
             (Ok(lhs_result), Ok(rhs_result)) => DiffEqResult::Success(lhs_result, rhs_result),
 
-            // Both sides failed. Check that the trap and state at the time of
-            // failure is the same, when possible.
+            // Handle all non-deterministic errors by poisoning this execution's
+            // state, so that we simply move on to the next test.
+            (Err(lhs), _) if lhs_engine.is_non_deterministic_error(&lhs) => {
+                log::debug!("lhs failed non-deterministically: {lhs:?}");
+                DiffEqResult::Poisoned
+            }
+            (_, Err(rhs)) if wasmtime_error_is_non_deterministic(&rhs) => {
+                log::debug!("rhs failed non-deterministically: {rhs:?}");
+                DiffEqResult::Poisoned
+            }
+
+            // Both sides failed deterministically. Check that the trap and
+            // state at the time of failure is the same.
             (Err(lhs), Err(rhs)) => {
-                let rhs = match rhs.downcast::<Trap>() {
-                    Ok(trap) => trap,
+                let rhs = rhs
+                    .downcast::<Trap>()
+                    .expect("non-traps handled in earlier match arm");
 
-                    // For general, unknown errors, we can't rely on this being
-                    // a deterministic Wasm failure that both engines handled
-                    // identically, leaving Wasm in identical states. We could
-                    // just as easily be hitting engine-specific failures, like
-                    // different implementation-defined limits. So simply report
-                    // failure and move on to the next test.
-                    Err(err) => {
-                        log::debug!("rhs failed: {err:?}");
-                        return DiffEqResult::Failed;
-                    }
-                };
-
-                // Even some traps are nondeterministic, and we can't rely on
-                // the errors matching or leaving Wasm in the same state.
-                let poisoned =
-                    // Allocations being too large for the GC are
-                    // implementation-defined.
-                    rhs == Trap::AllocationTooLarge
-                    // Stack size, and therefore when overflow happens, is
-                    // implementation-defined.
-                    || rhs == Trap::StackOverflow
-                    || lhs_engine.is_stack_overflow(&lhs);
-                if poisoned {
-                    return DiffEqResult::Poisoned;
-                }
+                debug_assert!(
+                    !lhs_engine.is_non_deterministic_error(&lhs),
+                    "non-deterministic traps handled in earlier match arm",
+                );
+                debug_assert!(
+                    !wasmtime_trap_is_non_deterministic(&rhs),
+                    "non-deterministic traps handled in earlier match arm",
+                );
 
                 lhs_engine.assert_error_match(&lhs, &rhs);
                 DiffEqResult::Failed
             }
+
             // A real bug is found if only one side fails.
             (Ok(_), Err(err)) => panic!("only the `rhs` failed for this input: {err:?}"),
             (Err(err), Ok(_)) => panic!("only the `lhs` failed for this input: {err:?}"),
@@ -642,7 +680,11 @@ pub fn make_api_calls(api: generators::api::ApiCalls) {
                 let nth = nth % funcs.len();
                 let f = &funcs[nth];
                 let ty = f.ty(&store);
-                if let Ok(params) = dummy::dummy_values(ty.params()) {
+                if let Some(params) = ty
+                    .params()
+                    .map(|p| p.default_value())
+                    .collect::<Option<Vec<_>>>()
+                {
                     let mut results = vec![Val::I32(0); ty.results().len()];
                     let _ = f.call(store, &params, &mut results);
                 }
@@ -654,22 +696,29 @@ pub fn make_api_calls(api: generators::api::ApiCalls) {
 /// Executes the wast `test` with the `config` specified.
 ///
 /// Ensures that wast tests pass regardless of the `Config`.
-pub fn wast_test(mut fuzz_config: generators::Config, test: generators::WastTest) {
+pub fn wast_test(u: &mut arbitrary::Unstructured<'_>) -> arbitrary::Result<()> {
     crate::init_fuzzing();
+
+    let mut fuzz_config: generators::Config = u.arbitrary()?;
+    let test: generators::WastTest = u.arbitrary()?;
+    if u.arbitrary()? {
+        fuzz_config.enable_async(u)?;
+    }
+
     let test = &test.test;
 
     // Discard tests that allocate a lot of memory as we don't want to OOM the
     // fuzzer and we also limit memory growth which would cause the test to
     // fail.
     if test.config.hogs_memory.unwrap_or(false) {
-        return;
+        return Err(arbitrary::Error::IncorrectFormat);
     }
 
     // Transform `fuzz_config` to be valid for `test` and make sure that this
     // test is supposed to pass.
     let wast_config = fuzz_config.make_wast_test_compliant(test);
     if test.should_fail(&wast_config) {
-        return;
+        return Err(arbitrary::Error::IncorrectFormat);
     }
 
     // Winch requires AVX and AVX2 for SIMD tests to pass so don't run the test
@@ -688,19 +737,24 @@ pub fn wast_test(mut fuzz_config: generators::Config, test: generators::WastTest
         log::warn!(
             "Skipping Wast test because Winch doesn't support SIMD tests with AVX or AVX2 disabled"
         );
-        return;
+        return Err(arbitrary::Error::IncorrectFormat);
     }
 
     // Fuel and epochs don't play well with threads right now, so exclude any
     // thread-spawning test if it looks like threads are spawned in that case.
     if fuzz_config.wasmtime.consume_fuel || fuzz_config.wasmtime.epoch_interruption {
         if test.contents.contains("(thread") {
-            return;
+            return Err(arbitrary::Error::IncorrectFormat);
         }
     }
 
     log::debug!("running {:?}", test.path);
-    let mut wast_context = WastContext::new(fuzz_config.to_store());
+    let async_ = if fuzz_config.wasmtime.async_config == generators::AsyncConfig::Disabled {
+        wasmtime_wast::Async::No
+    } else {
+        wasmtime_wast::Async::Yes
+    };
+    let mut wast_context = WastContext::new(fuzz_config.to_store(), async_);
     wast_context
         .register_spectest(&wasmtime_wast::SpectestConfig {
             use_shared_memory: true,
@@ -710,6 +764,7 @@ pub fn wast_test(mut fuzz_config: generators::Config, test: generators::WastTest
     wast_context
         .run_buffer(test.path.to_str().unwrap(), test.contents.as_bytes())
         .unwrap();
+    Ok(())
 }
 
 /// Execute a series of `table.get` and `table.set` operations.
@@ -954,9 +1009,11 @@ impl Drop for HelperThread {
 /// arbitrary types and values.
 pub fn dynamic_component_api_target(input: &mut arbitrary::Unstructured) -> arbitrary::Result<()> {
     use crate::generators::component_types;
-    use component_fuzz_util::{TestCase, Type, EXPORT_FUNCTION, IMPORT_FUNCTION, MAX_TYPE_DEPTH};
-    use component_test_util::FuncExt;
     use wasmtime::component::{Component, Linker, Val};
+    use wasmtime_test_util::component::FuncExt;
+    use wasmtime_test_util::component_fuzz::{
+        TestCase, Type, EXPORT_FUNCTION, IMPORT_FUNCTION, MAX_TYPE_DEPTH,
+    };
 
     crate::init_fuzzing();
 
@@ -982,7 +1039,7 @@ pub fn dynamic_component_api_target(input: &mut arbitrary::Unstructured) -> arbi
         encoding2: input.arbitrary()?,
     };
 
-    let mut config = component_test_util::config();
+    let mut config = wasmtime_test_util::component::config();
     config.debug_adapter_modules(input.arbitrary()?);
     let engine = Engine::new(&config).unwrap();
     let mut store = Store::new(&engine, (Vec::new(), None));
@@ -1076,17 +1133,17 @@ pub fn call_async(wasm: &[u8], config: &generators::Config, mut poll_amts: &[u32
                         log::info!("yielding {} times in import", poll_amt);
                         YieldN(poll_amt).await;
                         for (ret_ty, result) in ty.results().zip(results) {
-                            *result = dummy::dummy_value(ret_ty)?;
+                            *result = ret_ty.default_value().unwrap();
                         }
                         Ok(())
                     })
                 })
                 .into()
             }
-            other_ty => match dummy::dummy_extern(&mut store, other_ty) {
-                Ok(item) => item,
-                Err(e) => {
-                    log::warn!("couldn't create import: {}", e);
+            other_ty => match other_ty.default_value(&mut store) {
+                Some(item) => item,
+                None => {
+                    log::warn!("couldn't create import for {import:?}");
                     return;
                 }
             },
@@ -1134,11 +1191,11 @@ pub fn call_async(wasm: &[u8], config: &generators::Config, mut poll_amts: &[u32
         let ty = func.ty(&store);
         let params = ty
             .params()
-            .map(|ty| dummy::dummy_value(ty).unwrap())
+            .map(|ty| ty.default_value().unwrap())
             .collect::<Vec<_>>();
         let mut results = ty
             .results()
-            .map(|ty| dummy::dummy_value(ty).unwrap())
+            .map(|ty| ty.default_value().unwrap())
             .collect::<Vec<_>>();
 
         log::info!("invoking export {:?}", name);
@@ -1268,6 +1325,22 @@ mod tests {
         false
     }
 
+    /// Runs `f` with random data until it returns `Ok(())` `iters` times.
+    fn test_n_times<T: for<'a> Arbitrary<'a>>(
+        iters: u32,
+        mut f: impl FnMut(T, &mut Unstructured<'_>) -> arbitrary::Result<()>,
+    ) {
+        let mut to_test = 0..iters;
+        let ok = gen_until_pass(|a, b| {
+            if f(a, b).is_ok() {
+                Ok(to_test.next().is_none())
+            } else {
+                Ok(false)
+            }
+        });
+        assert!(ok);
+    }
+
     // Test that the `table_ops` fuzzer eventually runs the gc function in the host.
     // We've historically had issues where this fuzzer accidentally wasn't fuzzing
     // anything for a long time so this is an attempt to prevent that from happening
@@ -1308,6 +1381,8 @@ mod tests {
             | WasmFeatures::TAIL_CALL
             | WasmFeatures::WIDE_ARITHMETIC
             | WasmFeatures::MEMORY64
+            | WasmFeatures::FUNCTION_REFERENCES
+            | WasmFeatures::GC
             | WasmFeatures::GC_TYPES
             | WasmFeatures::CUSTOM_PAGE_SIZES
             | WasmFeatures::EXTENDED_CONST;
@@ -1353,4 +1428,9 @@ mod tests {
             panic!("never generated wasm module using {expected:?}");
         }
     }
+
+    #[test]
+    fn wast_smoke_test() {
+        test_n_times(50, |(), u| super::wast_test(u));
+    }
 }
