@@ -7,6 +7,7 @@ use crate::runtime::type_registry::TypeRegistry;
 use crate::runtime::vm::GcRuntime;
 use crate::Config;
 use alloc::sync::Arc;
+use core::ptr::NonNull;
 #[cfg(target_has_atomic = "64")]
 use core::sync::atomic::{AtomicU64, Ordering};
 #[cfg(any(feature = "cranelift", feature = "winch"))]
@@ -225,13 +226,13 @@ impl Engine {
     /// [`Engine::precompile_module`], or [`Engine::precompile_component`], then
     /// this will return `Some(...)` indicating so. Otherwise `None` is
     /// returned.
-    pub fn detect_precompiled(&self, bytes: &[u8]) -> Option<Precompiled> {
+    pub fn detect_precompiled(bytes: &[u8]) -> Option<Precompiled> {
         serialization::detect_precompiled_bytes(bytes)
     }
 
     /// Like [`Engine::detect_precompiled`], but performs the detection on a file.
     #[cfg(feature = "std")]
-    pub fn detect_precompiled_file(&self, path: impl AsRef<Path>) -> Result<Option<Precompiled>> {
+    pub fn detect_precompiled_file(path: impl AsRef<Path>) -> Result<Option<Precompiled>> {
         serialization::detect_precompiled_file(path)
     }
 
@@ -391,6 +392,7 @@ impl Engine {
             | "regalloc_algorithm"
             | "is_pic"
             | "bb_padding_log2_minus_one"
+            | "log2_min_function_alignment"
             | "machine_code_cfg_info"
             | "tls_model" // wasmtime doesn't use tls right now
             | "stack_switch_model" // wasmtime doesn't use stack switching right now
@@ -783,6 +785,29 @@ impl Engine {
         )
     }
 
+    /// Loads a `CodeMemory` from the specified memory region without copying
+    ///
+    /// The `expected` marker here is whether the bytes are expected to be
+    /// a precompiled module or a component.  The `memory` provided is expected
+    /// to be a serialized module (.cwasm) generated by `[Module::serialize]`
+    /// or [`Engine::precompile_module] or their `Component` counterparts
+    /// [`Component::serialize`] or `[Engine::precompile_component]`.
+    ///
+    /// The memory provided is guaranteed to only be immutably by the runtime.
+    ///
+    /// # Safety
+    ///
+    /// As there is no copy here, the runtime will be making direct readonly use
+    /// of the provided memory. As such, outside writes to this memory region
+    /// will result in undefined and likely very undesirable behavior.
+    pub(crate) unsafe fn load_code_raw(
+        &self,
+        memory: NonNull<[u8]>,
+        expected: ObjectKind,
+    ) -> Result<Arc<crate::CodeMemory>> {
+        self.load_code(crate::runtime::vm::MmapVec::from_raw(memory)?, expected)
+    }
+
     /// Like `load_code_bytes`, but creates a mmap from a file on disk.
     #[cfg(feature = "std")]
     pub(crate) fn load_code_file(
