@@ -8,8 +8,9 @@
 
 use crate::prelude::*;
 use crate::runtime::vm::{
-    SendSyncPtr, VMArrayCallFunction, VMFuncRef, VMGlobalDefinition, VMMemoryDefinition,
-    VMOpaqueContext, VMStore, VMStoreRawPtr, VMWasmCallFunction, ValRaw, VmPtr, VmSafe,
+    SendSyncPtr, VMArrayCallFunction, VMContext, VMFuncRef, VMGlobalDefinition, VMMemoryDefinition,
+    VMOpaqueContext, VMStore, VMStoreRawPtr, VMTable, VMTableDefinition, VMWasmCallFunction,
+    ValRaw, VmPtr, VmSafe,
 };
 use alloc::alloc::Layout;
 use alloc::sync::Arc;
@@ -86,6 +87,8 @@ pub struct ComponentInstance {
 ///   which this function pointer was registered.
 /// * `ty` - the type index, relative to the tables in `vmctx`, that is the
 ///   type of the function being called.
+/// * `caller_instance` - the `RuntimeComponentInstanceIndex` representing the
+///   caller component instance, used to track the owner of an async host task.
 /// * `flags` - the component flags for may_enter/leave corresponding to the
 ///   component instance that the lowering happened within.
 /// * `opt_memory` - this nullable pointer represents the memory configuration
@@ -106,7 +109,7 @@ pub struct ComponentInstance {
 /// or not. On failure this function records trap information in TLS which
 /// should be suitable for reading later.
 //
-// FIXME: 9 arguments is probably too many. The `data` through `string-encoding`
+// FIXME: 11 arguments is probably too many. The `data` through `string-encoding`
 // parameters should probably get packaged up into the `VMComponentContext`.
 // Needs benchmarking one way or another though to figure out what the best
 // balance is here.
@@ -114,6 +117,7 @@ pub type VMLoweringCallee = extern "C" fn(
     vmctx: NonNull<VMOpaqueContext>,
     data: NonNull<u8>,
     ty: u32,
+    caller_instance: u32,
     flags: NonNull<VMGlobalDefinition>,
     opt_memory: *mut VMMemoryDefinition,
     opt_realloc: *mut VMFuncRef,
@@ -286,6 +290,20 @@ impl ComponentInstance {
         }
     }
 
+    /// Returns the runtime table definition and associated instance `VMContext`
+    /// corresponding to the index of the table provided.
+    ///
+    /// This can only be called after `idx` has been initialized at runtime
+    /// during the instantiation process of a component.
+    pub fn runtime_table(&self, idx: RuntimeTableIndex) -> VMTable {
+        unsafe {
+            let ret = *self.vmctx_plus_offset::<VMTable>(self.offsets.runtime_table(idx));
+            debug_assert!(ret.from.as_ptr() as usize != INVALID_PTR);
+            debug_assert!(ret.vmctx.as_ptr() as usize != INVALID_PTR);
+            ret
+        }
+    }
+
     /// Returns the realloc pointer corresponding to the index provided.
     ///
     /// This can only be called after `idx` has been initialized at runtime
@@ -401,6 +419,31 @@ impl ComponentInstance {
         }
     }
 
+    /// Stores the runtime table pointer at the index specified.
+    ///
+    /// This is intended to be called during the instantiation process of a
+    /// component once a table is available, which may not be until part-way
+    /// through component instantiation.
+    ///
+    /// Note that it should be a property of the component model that the `ptr`
+    /// here is never needed prior to it being configured here in the instance.
+    pub fn set_runtime_table(
+        &mut self,
+        idx: RuntimeTableIndex,
+        ptr: NonNull<VMTableDefinition>,
+        vmctx: NonNull<VMContext>,
+    ) {
+        unsafe {
+            let storage = self.vmctx_plus_offset_mut::<VMTable>(self.offsets.runtime_table(idx));
+            debug_assert!((*storage).vmctx.as_ptr() as usize == INVALID_PTR);
+            debug_assert!((*storage).from.as_ptr() as usize == INVALID_PTR);
+            *storage = VMTable {
+                vmctx: vmctx.into(),
+                from: ptr.into(),
+            };
+        }
+    }
+
     /// Configures host runtime lowering information associated with imported f
     /// functions for the `idx` specified.
     pub fn set_lowering(&mut self, idx: LoweredIndex, lowering: VMLowering) {
@@ -520,6 +563,11 @@ impl ComponentInstance {
                 let offset = self.offsets.resource_destructor(i);
                 *self.vmctx_plus_offset_mut(offset) = INVALID_PTR;
             }
+            for i in 0..self.offsets.num_runtime_tables {
+                let i = RuntimeTableIndex::from_u32(i);
+                let offset = self.offsets.runtime_table(i);
+                *self.vmctx_plus_offset_mut(offset) = INVALID_PTR;
+            }
         }
     }
 
@@ -664,6 +712,39 @@ impl ComponentInstance {
     pub(crate) fn resource_exit_call(&mut self) -> Result<()> {
         self.resource_tables().exit_call()
     }
+
+    #[cfg(feature = "component-model-async")]
+    pub(crate) fn future_transfer(
+        &mut self,
+        src_idx: u32,
+        src: TypeFutureTableIndex,
+        dst: TypeFutureTableIndex,
+    ) -> Result<u32> {
+        _ = (src_idx, src, dst);
+        todo!()
+    }
+
+    #[cfg(feature = "component-model-async")]
+    pub(crate) fn stream_transfer(
+        &mut self,
+        src_idx: u32,
+        src: TypeStreamTableIndex,
+        dst: TypeStreamTableIndex,
+    ) -> Result<u32> {
+        _ = (src_idx, src, dst);
+        todo!()
+    }
+
+    #[cfg(feature = "component-model-async")]
+    pub(crate) fn error_context_transfer(
+        &mut self,
+        src_idx: u32,
+        src: TypeComponentLocalErrorContextTableIndex,
+        dst: TypeComponentLocalErrorContextTableIndex,
+    ) -> Result<u32> {
+        _ = (src_idx, src, dst);
+        todo!()
+    }
 }
 
 impl VMComponentContext {
@@ -765,6 +846,16 @@ impl OwnedComponentInstance {
         unsafe { self.instance_mut().set_runtime_post_return(idx, ptr) }
     }
 
+    /// See `ComponentInstance::set_runtime_table`
+    pub fn set_runtime_table(
+        &mut self,
+        idx: RuntimeTableIndex,
+        ptr: NonNull<VMTableDefinition>,
+        vmctx: NonNull<VMContext>,
+    ) {
+        unsafe { self.instance_mut().set_runtime_table(idx, ptr, vmctx) }
+    }
+
     /// See `ComponentInstance::set_lowering`
     pub fn set_lowering(&mut self, idx: LoweredIndex, lowering: VMLowering) {
         unsafe { self.instance_mut().set_lowering(idx, lowering) }
