@@ -144,7 +144,7 @@ mod test_vmfunction_body {
 /// imported from another instance.
 #[derive(Debug, Copy, Clone)]
 #[repr(C)]
-pub struct VMTableImport {
+pub struct VMTable {
     /// A pointer to the imported table description.
     pub from: VmPtr<VMTableDefinition>,
 
@@ -153,30 +153,43 @@ pub struct VMTableImport {
 }
 
 // SAFETY: the above structure is repr(C) and only contains `VmSafe` fields.
-unsafe impl VmSafe for VMTableImport {}
+unsafe impl VmSafe for VMTable {}
 
 #[cfg(test)]
-mod test_vmtable_import {
-    use super::VMTableImport;
+mod test_vmtable {
+    use super::VMTable;
     use core::mem::offset_of;
     use std::mem::size_of;
+    use wasmtime_environ::component::{Component, VMComponentOffsets};
     use wasmtime_environ::{HostPtr, Module, VMOffsets};
 
     #[test]
-    fn check_vmtable_import_offsets() {
+    fn check_vmtable_offsets() {
         let module = Module::new();
         let offsets = VMOffsets::new(HostPtr, &module);
+        assert_eq!(size_of::<VMTable>(), usize::from(offsets.size_of_vmtable()));
         assert_eq!(
-            size_of::<VMTableImport>(),
-            usize::from(offsets.size_of_vmtable_import())
+            offset_of!(VMTable, from),
+            usize::from(offsets.vmtable_from())
         );
         assert_eq!(
-            offset_of!(VMTableImport, from),
-            usize::from(offsets.vmtable_import_from())
+            offset_of!(VMTable, vmctx),
+            usize::from(offsets.vmtable_vmctx())
         );
+    }
+
+    #[test]
+    fn ensure_sizes_match() {
+        // Because we use `VMTable` for recording tables used by components, we
+        // want to make sure that the size calculations between `VMOffsets` and
+        // `VMComponentOffsets` stay the same.
+        let module = Module::new();
+        let vm_offsets = VMOffsets::new(HostPtr, &module);
+        let component = Component::default();
+        let vm_component_offsets = VMComponentOffsets::new(HostPtr, &component);
         assert_eq!(
-            offset_of!(VMTableImport, vmctx),
-            usize::from(offsets.vmtable_import_vmctx())
+            vm_offsets.size_of_vmtable(),
+            vm_component_offsets.size_of_vmtable()
         );
     }
 }
@@ -947,7 +960,6 @@ macro_rules! define_builtin_array {
             /// pointer is considered valid.
             pub fn expose_provenance(&self) -> NonNull<Self>{
                 $(
-                    #[cfg(has_provenance_apis)]
                     (self.$name as *mut u8).expose_provenance();
                 )*
                 NonNull::from(self)
