@@ -2,15 +2,18 @@ use crate::common::{Profile, RunCommon, RunTarget};
 use anyhow::{anyhow, bail, Result};
 use clap::Parser;
 use std::net::SocketAddr;
+use std::time::Instant;
 use std::{
     path::PathBuf,
     sync::{
         atomic::{AtomicBool, AtomicU64, Ordering},
-        Arc,
+        Arc, Mutex,
     },
+    time::Duration,
 };
-use wasmtime::component::Linker;
-use wasmtime::{Engine, Store, StoreLimits};
+use tokio::sync::Notify;
+use wasmtime::component::{Component, Linker};
+use wasmtime::{Engine, Store, StoreLimits, UpdateDeadline};
 use wasmtime_wasi::{IoView, StreamError, StreamResult, WasiCtx, WasiCtxBuilder, WasiView};
 use wasmtime_wasi_http::bindings::http::types::Scheme;
 use wasmtime_wasi_http::bindings::ProxyPre;
@@ -44,6 +47,9 @@ struct Host {
 
     #[cfg(feature = "wasi-keyvalue")]
     wasi_keyvalue: Option<WasiKeyValueCtx>,
+
+    #[cfg(feature = "profiling")]
+    guest_profiler: Option<Arc<wasmtime::GuestProfiler>>,
 }
 
 impl IoView for Host {
@@ -85,12 +91,19 @@ pub struct ServeCommand {
     run: RunCommon,
 
     /// Socket address for the web server to bind to.
-    #[arg(long = "addr", value_name = "SOCKADDR", default_value_t = DEFAULT_ADDR)]
+    #[arg(long , value_name = "SOCKADDR", default_value_t = DEFAULT_ADDR)]
     addr: SocketAddr,
 
+    /// Socket address where, when connected to, will initiate a graceful
+    /// shutdown.
+    ///
+    /// Note that graceful shutdown is also supported on ctrl-c.
+    #[arg(long, value_name = "SOCKADDR")]
+    shutdown_addr: Option<SocketAddr>,
+
     /// Disable log prefixes of wasi-http handlers.
     /// if unspecified, logs will be prefixed with 'stdout|stderr [{req_id}] :: '
-    #[arg(long = "no-logging-prefix")]
+    #[arg(long)]
     no_logging_prefix: bool,
 
     /// The WebAssembly component to run.
@@ -105,9 +118,6 @@ impl ServeCommand {
 
         // We force cli errors before starting to listen for connections so then
         // we don't accidentally delay them to the first request.
-        if let Some(Profile::Guest { .. }) = &self.run.profile {
-            bail!("Cannot use the guest profiler with components");
-        }
 
         if self.run.common.wasi.nn == Some(true) {
             #[cfg(not(feature = "wasi-nn"))]
@@ -134,17 +144,7 @@ impl ServeCommand {
             .enable_io()
             .build()?;
 
-        runtime.block_on(async move {
-            tokio::select! {
-                _ = tokio::signal::ctrl_c() => {
-                    Ok::<_, anyhow::Error>(())
-                }
-
-                res = self.serve() => {
-                    res
-                }
-            }
-        })?;
+        runtime.block_on(self.serve())?;
 
         Ok(())
     }
@@ -182,6 +182,8 @@ impl ServeCommand {
             wasi_config: None,
             #[cfg(feature = "wasi-keyvalue")]
             wasi_keyvalue: None,
+            #[cfg(feature = "profiling")]
+            guest_profiler: None,
         };
 
         if self.run.common.wasi.nn == Some(true) {
@@ -234,10 +236,6 @@ impl ServeCommand {
 
         let mut store = Store::new(engine, host);
 
-        if self.run.common.wasm.timeout.is_some() {
-            store.set_epoch_deadline(u64::from(EPOCH_PRECISION) + 1);
-        }
-
         store.data_mut().limits = self.run.store_limits();
         store.limiter(|t| &mut t.limits);
 
@@ -351,10 +349,9 @@ impl ServeCommand {
             Some(Profile::Native(s)) => {
                 config.profiler(s);
             }
-
-            // We bail early in `execute` if the guest profiler is configured.
-            Some(Profile::Guest { .. }) => unreachable!(),
-
+            Some(Profile::Guest { .. }) => {
+                config.epoch_interruption(true);
+            }
             None => {}
         }
 
@@ -371,6 +368,30 @@ impl ServeCommand {
         let instance = linker.instantiate_pre(&component)?;
         let instance = ProxyPre::new(instance)?;
 
+        // Spawn background task(s) waiting for graceful shutdown signals. This
+        // always listens for ctrl-c but additionally can listen for a TCP
+        // connection to the specified address.
+        let shutdown = Arc::new(GracefulShutdown::default());
+        tokio::task::spawn({
+            let shutdown = shutdown.clone();
+            async move {
+                tokio::signal::ctrl_c().await.unwrap();
+                shutdown.requested.notify_one();
+            }
+        });
+        if let Some(addr) = self.shutdown_addr {
+            let listener = tokio::net::TcpListener::bind(addr).await?;
+            eprintln!(
+                "Listening for shutdown on tcp://{}/",
+                listener.local_addr()?
+            );
+            let shutdown = shutdown.clone();
+            tokio::task::spawn(async move {
+                let _ = listener.accept().await;
+                shutdown.requested.notify_one();
+            });
+        }
+
         let socket = match &self.addr {
             SocketAddr::V4(_) => tokio::net::TcpSocket::new_v4()?,
             SocketAddr::V6(_) => tokio::net::TcpSocket::new_v6()?,
@@ -389,44 +410,111 @@ impl ServeCommand {
 
         eprintln!("Serving HTTP on http://{}/", listener.local_addr()?);
 
-        let _epoch_thread = if let Some(timeout) = self.run.common.wasm.timeout {
-            Some(EpochThread::spawn(
-                timeout / EPOCH_PRECISION,
-                engine.clone(),
-            ))
-        } else {
-            None
-        };
-
         log::info!("Listening on {}", self.addr);
 
         let handler = ProxyHandler::new(self, engine, instance);
 
         loop {
-            let (stream, _) = listener.accept().await?;
+            // Wait for a socket, but also "race" against shutdown to break out
+            // of this loop. Once the graceful shutdown signal is received then
+            // this loop exits immediately.
+            let (stream, _) = tokio::select! {
+                _ = shutdown.requested.notified() => break,
+                v = listener.accept() => v?,
+            };
+            let comp = component.clone();
             let stream = TokioIo::new(stream);
             let h = handler.clone();
-            tokio::task::spawn(async {
+            let shutdown_guard = shutdown.clone().increment();
+            tokio::task::spawn(async move {
                 if let Err(e) = http1::Builder::new()
                     .keep_alive(true)
                     .serve_connection(
                         stream,
-                        hyper::service::service_fn(move |req| handle_request(h.clone(), req)),
+                        hyper::service::service_fn(move |req| {
+                            handle_request(h.clone(), req, comp.clone())
+                        }),
                     )
                     .await
                 {
                     eprintln!("error: {e:?}");
                 }
+                drop(shutdown_guard);
             });
         }
+
+        // Upon exiting the loop we'll no longer process any more incoming
+        // connections but there may still be outstanding connections
+        // processing in child tasks. If there are wait for those to complete
+        // before shutting down completely. Also enable short-circuiting this
+        // wait with a second ctrl-c signal.
+        if shutdown.close() {
+            return Ok(());
+        }
+        eprintln!("Waiting for child tasks to exit, ctrl-c again to quit sooner...");
+        tokio::select! {
+            _ = tokio::signal::ctrl_c() => {}
+            _ = shutdown.complete.notified() => {}
+        }
+
+        Ok(())
     }
 }
 
-/// This is the number of epochs that we will observe before expiring a request handler. As
-/// instances may be started at any point within an epoch, and epochs are counted globally per
-/// engine, we expire after `EPOCH_PRECISION + 1` epochs have been observed. This gives a maximum
-/// overshoot of `timeout / EPOCH_PRECISION`, which is more desirable than expiring early.
-const EPOCH_PRECISION: u32 = 10;
+/// Helper structure to manage graceful shutdown int he accept loop above.
+#[derive(Default)]
+struct GracefulShutdown {
+    /// Async notification that shutdown has been requested.
+    requested: Notify,
+    /// Async notification that shutdown has completed, signaled when
+    /// `notify_when_done` is `true` and `active_tasks` reaches 0.
+    complete: Notify,
+    /// Internal state related to what's in progress when shutdown is requested.
+    state: Mutex<GracefulShutdownState>,
+}
+
+#[derive(Default)]
+struct GracefulShutdownState {
+    active_tasks: u32,
+    notify_when_done: bool,
+}
+
+impl GracefulShutdown {
+    /// Increments the number of active tasks and returns a guard indicating
+    fn increment(self: Arc<Self>) -> impl Drop {
+        struct Guard(Arc<GracefulShutdown>);
+
+        let mut state = self.state.lock().unwrap();
+        assert!(!state.notify_when_done);
+        state.active_tasks += 1;
+        drop(state);
+
+        return Guard(self);
+
+        impl Drop for Guard {
+            fn drop(&mut self) {
+                let mut state = self.0.state.lock().unwrap();
+                state.active_tasks -= 1;
+                if state.notify_when_done && state.active_tasks == 0 {
+                    self.0.complete.notify_one();
+                }
+            }
+        }
+    }
+
+    /// Flags this state as done spawning tasks and returns whether there are no
+    /// more child tasks remaining.
+    fn close(&self) -> bool {
+        let mut state = self.state.lock().unwrap();
+        state.notify_when_done = true;
+        state.active_tasks == 0
+    }
+}
+
+/// When executing with a timeout enabled, this is how frequently epoch
+/// interrupts will be executed to check for timeouts. If guest profiling
+/// is enabled, the guest epoch period will be used.
+const EPOCH_INTERRUPT_PERIOD: Duration = Duration::from_millis(50);
 
 struct EpochThread {
     shutdown: Arc<AtomicBool>,
@@ -434,13 +522,13 @@ struct EpochThread {
 }
 
 impl EpochThread {
-    fn spawn(timeout: std::time::Duration, engine: Engine) -> Self {
+    fn spawn(interval: std::time::Duration, engine: Engine) -> Self {
         let shutdown = Arc::new(AtomicBool::new(false));
         let handle = {
             let shutdown = Arc::clone(&shutdown);
             let handle = std::thread::spawn(move || {
                 while !shutdown.load(Ordering::Relaxed) {
-                    std::thread::sleep(timeout);
+                    std::thread::sleep(interval);
                     engine.increment_epoch();
                 }
             });
@@ -460,6 +548,121 @@ impl Drop for EpochThread {
     }
 }
 
+type WriteProfile = Box<dyn FnOnce(&mut Store<Host>) + Send>;
+
+fn setup_epoch_handler(
+    cmd: &ServeCommand,
+    store: &mut Store<Host>,
+    component: Component,
+) -> Result<(WriteProfile, Option<EpochThread>)> {
+    // Profiling Enabled
+    if let Some(Profile::Guest { interval, path }) = &cmd.run.profile {
+        #[cfg(feature = "profiling")]
+        return setup_guest_profiler(cmd, store, path.clone(), *interval, component.clone());
+        #[cfg(not(feature = "profiling"))]
+        {
+            let _ = (path, interval);
+            bail!("support for profiling disabled at compile time!");
+        }
+    }
+
+    // Profiling disabled but there's a global request timeout
+    let epoch_thread = if let Some(timeout) = cmd.run.common.wasm.timeout {
+        let start = Instant::now();
+        store.epoch_deadline_callback(move |_store| {
+            if start.elapsed() > timeout {
+                bail!("Timeout expired");
+            }
+            Ok(UpdateDeadline::Continue(1))
+        });
+        store.set_epoch_deadline(1);
+        let engine = store.engine().clone();
+        Some(EpochThread::spawn(EPOCH_INTERRUPT_PERIOD, engine))
+    } else {
+        None
+    };
+
+    Ok((Box::new(|_store| {}), epoch_thread))
+}
+
+#[cfg(feature = "profiling")]
+fn setup_guest_profiler(
+    cmd: &ServeCommand,
+    store: &mut Store<Host>,
+    path: String,
+    interval: Duration,
+    component: Component,
+) -> Result<(WriteProfile, Option<EpochThread>)> {
+    use wasmtime::{AsContext, GuestProfiler, StoreContext, StoreContextMut};
+
+    let module_name = "<main>";
+
+    store.data_mut().guest_profiler = Some(Arc::new(GuestProfiler::new_component(
+        module_name,
+        interval,
+        component,
+        std::iter::empty(),
+    )));
+
+    fn sample(
+        mut store: StoreContextMut<Host>,
+        f: impl FnOnce(&mut GuestProfiler, StoreContext<Host>),
+    ) {
+        let mut profiler = store.data_mut().guest_profiler.take().unwrap();
+        f(
+            Arc::get_mut(&mut profiler).expect("profiling doesn't support threads yet"),
+            store.as_context(),
+        );
+        store.data_mut().guest_profiler = Some(profiler);
+    }
+
+    // Hostcall entry/exit, etc.
+    store.call_hook(|store, kind| {
+        sample(store, |profiler, store| profiler.call_hook(store, kind));
+        Ok(())
+    });
+
+    let start = Instant::now();
+    let timeout = cmd.run.common.wasm.timeout;
+    store.epoch_deadline_callback(move |store| {
+        sample(store, |profiler, store| {
+            profiler.sample(store, std::time::Duration::ZERO)
+        });
+
+        // Originally epoch counting was used here; this is problematic in
+        // a lot of cases due to there being a lot of time (e.g. in hostcalls)
+        // when we are not expected to get sample hits.
+        if let Some(timeout) = timeout {
+            if start.elapsed() > timeout {
+                bail!("Timeout expired");
+            }
+        }
+
+        Ok(UpdateDeadline::Continue(1))
+    });
+
+    store.set_epoch_deadline(1);
+    let engine = store.engine().clone();
+    let epoch_thread = Some(EpochThread::spawn(interval, engine));
+
+    let write_profile = Box::new(move |store: &mut Store<Host>| {
+        let profiler = Arc::try_unwrap(store.data_mut().guest_profiler.take().unwrap())
+            .expect("profiling doesn't support threads yet");
+        if let Err(e) = std::fs::File::create(&path)
+            .map_err(anyhow::Error::new)
+            .and_then(|output| profiler.finish(std::io::BufWriter::new(output)))
+        {
+            eprintln!("failed writing profile at {path}: {e:#}");
+        } else {
+            eprintln!();
+            eprintln!("Profile written to: {path}");
+            eprintln!("View this profile at https://profiler.firefox.com/.");
+        }
+    });
+
+    Ok((write_profile, epoch_thread))
+}
+
 struct ProxyHandlerInner {
     cmd: ServeCommand,
     engine: Engine,
@@ -492,6 +695,7 @@ type Request = hyper::Request<hyper::body::Incoming>;
 async fn handle_request(
     ProxyHandler(inner): ProxyHandler,
     req: Request,
+    component: Component,
 ) -> Result<hyper::Response<HyperOutgoingBody>> {
     let (sender, receiver) = tokio::sync::oneshot::channel();
 
@@ -509,20 +713,26 @@ async fn handle_request(
     let out = store.data_mut().new_response_outparam(sender)?;
     let proxy = inner.instance_pre.instantiate_async(&mut store).await?;
 
+    let comp = component.clone();
     let task = tokio::task::spawn(async move {
+        let (write_profile, epoch_thread) = setup_epoch_handler(&inner.cmd, &mut store, comp)?;
+
         if let Err(e) = proxy
             .wasi_http_incoming_handler()
-            .call_handle(store, req, out)
+            .call_handle(&mut store, req, out)
             .await
         {
             log::error!("[{req_id}] :: {:?}", e);
             return Err(e);
         }
 
+        write_profile(&mut store);
+        drop(epoch_thread);
+
         Ok(())
     });
 
-    match receiver.await {
+    let result = match receiver.await {
         Ok(Ok(resp)) => Ok(resp),
         Ok(Err(e)) => Err(e.into()),
         Err(_) => {
@@ -534,12 +744,17 @@ async fn handle_request(
             // that we assume the task has already exited at this point so the
             // `await` should resolve immediately.
             let e = match task.await {
-                Ok(r) => r.expect_err("if the receiver has an error, the task must have failed"),
+                Ok(Ok(())) => {
+                    bail!("guest never invoked `response-outparam::set` method")
+                }
+                Ok(Err(e)) => e,
                 Err(e) => e.into(),
             };
-            return Err(e.context("guest never invoked `response-outparam::set` method"));
+            Err(e.context("guest never invoked `response-outparam::set` method"))
         }
-    }
+    };
+
+    result
 }
 
 #[derive(Clone)]
