@@ -1143,10 +1143,11 @@ where
     /// Emits a series of instructions that load the `fuel_consumed` field from
     /// `VMStoreContext`.
     fn emit_load_fuel_consumed(&mut self, fuel_reg: Reg) -> Result<()> {
-        let limits_offset = self.env.vmoffsets.ptr.vmctx_runtime_limits();
+        let store_context_offset = self.env.vmoffsets.ptr.vmctx_store_context();
         let fuel_offset = self.env.vmoffsets.ptr.vmstore_context_fuel_consumed();
         self.masm.load_ptr(
-            self.masm.address_at_vmctx(u32::from(limits_offset))?,
+            self.masm
+                .address_at_vmctx(u32::from(store_context_offset))?,
             writable!(fuel_reg),
         )?;
 
@@ -1221,7 +1222,7 @@ where
         epoch_counter_reg: Reg,
     ) -> Result<()> {
         let epoch_ptr_offset = self.env.vmoffsets.ptr.vmctx_epoch_ptr();
-        let runtime_limits_offset = self.env.vmoffsets.ptr.vmctx_runtime_limits();
+        let store_context_offset = self.env.vmoffsets.ptr.vmctx_store_context();
         let epoch_deadline_offset = self.env.vmoffsets.ptr.vmstore_context_epoch_deadline();
 
         // Load the current epoch value into `epoch_counter_var`.
@@ -1241,7 +1242,7 @@ where
         // Load the `VMStoreContext`.
         self.masm.load_ptr(
             self.masm
-                .address_at_vmctx(u32::from(runtime_limits_offset))?,
+                .address_at_vmctx(u32::from(store_context_offset))?,
             writable!(epoch_deadline_reg),
         )?;
 
@@ -1262,13 +1263,14 @@ where
             return Ok(());
         }
 
-        let limits_offset = self.env.vmoffsets.ptr.vmctx_runtime_limits();
+        let store_context_offset = self.env.vmoffsets.ptr.vmctx_store_context();
         let fuel_offset = self.env.vmoffsets.ptr.vmstore_context_fuel_consumed();
         let limits_reg = self.context.any_gpr(self.masm)?;
 
         // Load `VMStoreContext` into the `limits_reg` reg.
         self.masm.load_ptr(
-            self.masm.address_at_vmctx(u32::from(limits_offset))?,
+            self.masm
+                .address_at_vmctx(u32::from(store_context_offset))?,
             writable!(limits_reg),
         )?;
 
